 module rungekutta  !! Runge-Kutta methods, 2D problems with one parameter
   implicit none
   integer,parameter,public :: db = kind(1.d0)
   integer,parameter,public :: ll = selected_real_kind(p=16)
   real(kind=db),parameter,public    :: hmax = 0.5_db, hmin = 0.00062_db,&
                                        hinit = 10._db*hmin, mins = 0.15_db, maxs = 10._db
   real(kind=db),parameter,private :: a21=1._db/5._db,&
                                      a31=3._db/40._db, a32= 9._db/40._db,&
                                      a41=44._db/45._db, a42= -56._db/15._db, a43= 32._db/9._db,&
                                      a51=19372._db/6561._db, a52=-25360/2817._db, a53=64448._db/6561._db, &
                                                              a54=-212._db/729._db,&
                                      a61=9017._db/3168._db, a62=-355._db/33._db, a63=46723._db/5247._db, &
                                                              a64=49._db/174._db, a65=-5103._db/18656._db,&
                                      a71=35._db/384._db, a73=500._db/1113._db, a74=125._db/192._db, &
                                                              a75=-2187._db/6784._db, a76= 11._db/84._db,&
                                      c2=1._db/5._db, c3=3._db/10._db, c4=4._db/5._db, c5= 8._db/9._db,&
                                      b1=5179._db/57600._db, b3=7571._db/16695._db, b4=393._db/640._db, &
                                                              b5=-92097._db/339200._db,&
                                      b6=187._db/2100._db, b7=1._db/40._db, &
                                      e1 = 71._db/57600._db, e3 = -71._db/16695._db, e4 = 71._db/1920._db, &
                                                              e5 = -17253._db/339200._db, e6 = 22._db/525._db
    integer,parameter,public :: passi = 30
   contains
   
   
   


   pure function RKDP_adaptive(f,x0,periodo,lambda,atoll,rtoll) result(y) 
       implicit none
       interface
         pure function f(t,lambda,xi)
           integer,parameter :: db = kind(1.d0)
           real(kind=db),intent(in) :: t,lambda
           real(kind=db),intent(in),dimension(2) :: xi
           real(kind=db),dimension(2) :: f
           end function f
         end interface
       real(kind=db),intent(IN),dimension(2) :: x0
       real(kind=db),intent(IN) :: lambda, periodo,atoll,rtoll
       real(kind=db),dimension(2) :: y
       real(kind=db) :: R, Rold, errass, s, h, t0
       real(kind=ll),dimension(2) :: k0, k1, k2, k3, k4, k5, k6, k7, k8, yy, yy_new
       
       Rold = atoll
       t0 = 0._db
       h = hinit
       yy = x0
       k0 =  f(t0,lambda,x0)
       
       mcycle: do !while ( t0 .lt. periodo )
         attempts: do
         k1 = k0
         k2 = f(t0+c2*h,lambda,real(yy+h*a21*k1,kind=db))
         k3 = f(t0+c3*h,lambda,real(yy+h*(a31*k1+a32*k2),kind=db))
         k4 = f(t0+c4*h,lambda,real(yy+h*(a41*k1+a42*k2+a43*k3),kind=db))
         k5 = f(t0+c5*h,lambda,real(yy+h*(a51*k1+a52*k2+a53*k3+a54*k4),kind=db))
         k6 = f(t0+h,lambda,real(yy+h*(a61*k1+a62*k2+a63*k3+a64*k4+a65*k5),kind=db))
         k7 = f(t0+h,lambda,real(yy+h*(a71*k1+a73*k3+a74*k4+a75*k5+a76*k6),kind=db))
         yy_new = yy+h*(k1*b1+k3*b3+k4*b3+k5*b5+k6*b6+k7*b7) !! 6th order approx
         !yy5 = yy+h*(k1*a71+k3*a73+k4*a74+k5*a75+k6*a76)    !! 5th order (not used)
         errass = real(h*maxval(abs(k1*e1+k3*e3+k4*e4+k5*e5+k6*e6-k7*b7)),kind=db)
         R = max(atoll,errass/(atoll+rtoll*max(maxval(abs(yy_new)),maxval(abs(yy)))))            
         s = max(mins,min(maxs,0.9_db*R**(-0.17_db)*(Rold**0.04_db)) ) !! PI  control: Alpha=0.2, Beta = 0.04
         
         if ((R .lt. 1.) .or. (h .le. hmin))  then  !! accept step
           t0= t0 + h
           k0 = k7
           yy=yy_new
           h=min(s*h,hmax,periodo-t0)
           Rold = R
           exit attempts
         else                                       !! reject step
           h=max(min(s,1._db)*h,hmin)
           k0 = f(t0,lambda,real(yy,kind=db))
           endif
           end do attempts
           
         if ( t0 .ge. periodo)   exit mcycle
         end do mcycle
         y = real(yy,kind=db)
       
       end function RKDP_adaptive





   

   end module rungekutta


   