!! Analyzes a 3D-grid, looking for initial points of periodic solutions. Search
!! is refined with a Newton iteration when promising points in the grid are found.
program disegno
  use service
  implicit none
  interface
    pure function deriv(t,lambda,xi)
      integer,parameter :: db = kind(1.d0)
      real(kind=db),intent(in) :: t,lambda
      real(kind=db),intent(in),dimension(2) :: xi
      real(kind=db),dimension(2) :: deriv
    end function deriv
  end interface
  integer :: i,j,k,punti, punti2, effp, effp2, nstepx,nstepy,nstepl, rp, gp
  real(kind=db),dimension(2) :: x0, y0
  real(kind=db) :: l,stepx,stepy,stepl, a0,a1,b0,b1,l0,l1
  real(kind=db),allocatable,dimension(:,:) :: punti_grezzi, punti_buoni
  logical :: success
  character (len=15) :: arg0

  call getarg(1,arg0)
  read(arg0,*) a0  !Left bound for x  
  call getarg(2,arg0)
  read(arg0,*) a1  !Right bound for x
  call getarg(3,arg0)
  read(arg0,*) b0  !Left bound for y  
  call getarg(4,arg0)
  read(arg0,*) b1  !Right bound for y
  call getarg(5,arg0)
  read(arg0,*) l0  !Smallest bound for parameter l
  call getarg(6,arg0)
  read(arg0,*) l1  !Largest bound for parameter l
  call getarg(7,arg0)
  read(arg0,*) nstepx  !number of x-intervals
  call getarg(8,arg0)
  read(arg0,*) nstepy  !number of y-intervals
  call getarg(9,arg0)
  read(arg0,*) nstepl  !number of l-intervals
  call getarg(10,arg0)
  read(arg0,*) tempo  !time for time translation
  call getarg(11,arg0)
  read(arg0,*) tol   ! what we really consider a zero
  call getarg(12,arg0)
  read(arg0,*) atoll   ! absolute tollerance for rkdp
  call getarg(13,arg0)
  read(arg0,*) rtoll   ! relative tollerance for rkdp
  call getarg(14,arg0)
  read(arg0,*) rp     ! maximum # of rough points
  call getarg(15,arg0)
  read(arg0,*) gp     ! maximum number of good points
  
  rp = rp +1
  gp = gp +1
  
  allocate(punti_grezzi(3,rp))
  allocate(punti_buoni(3,gp))

  stepx = (a1-a0)/nstepx
  stepy = (b1-b0)/nstepy
  stepl = (l1-l0)/nstepl
  
  maxst = overlap* (/ stepx , stepy /)

  punti = 0
  punti2 = 0
  effp = 0 
  effp2 = 0


!$OMP parallel PRIVATE(i,j,k,l,x0,y0,success)&
!$OMP          SHARED(stepx,stepy,stepl,tempo,l0,a0,b0,maxst,punti,punti2,rp,gp,effp,effp2, &
!$OMP               nstepl,nstepx,nstepy,atoll,rtoll,tol,punti_grezzi,punti_buoni)&
!$OMP          DEFAULT(NONE)
!$OMP do SCHEDULE(dynamic)  !! do it forward
  do i=0,nstepl
    l=i*stepl+l0
    do j=0,nstepx
      x0(1)=stepx*j+a0
      do k=0,nstepy
        x0(2)=stepy*k+b0
        y0 = ff(x0,l)
        if ( all(abs(y0) .lt. maxst) ) then
!$OMP CRITICAL      
          punti2= min(punti2 +1,rp)
          effp2 = effp2 +1
          punti_grezzi(:,punti2)= (/ x0,l /) 
!$OMP END CRITICAL
          call newtit(ff,l,x0,y0,success)
          if (success) then
!$OMP CRITICAL     
            punti = min(punti + 1,gp)
            effp = effp +1
            punti_buoni(:,punti)= (/ x0,l /)
!$OMP END CRITICAL
            endif
          endif
      end do
    end do
  end do
!$OMP end do nowait
!$OMP do SCHEDULE(dynamic)  !! do it backward
  do i=0,nstepl
    l=i*stepl+l0
    do j=0,nstepx
      x0(1)=stepx*j+a0
      do k=0,nstepy
        x0(2)=stepy*k+b0
        y0 = fb(x0,l)
        if ( all(abs(y0) .lt. maxst) ) then
!$OMP CRITICAL      
          punti2 = min(punti2 +1,rp)
          effp2 = effp2 +1
          punti_grezzi(:,punti2)= (/ x0,l /)
!$OMP END CRITICAL
          call newtit(fb,l,x0,y0,success)
          if (success) then
!$OMP CRITICAL     
            punti = min(punti + 1,gp)
            effp = effp +1
            punti_buoni(:,punti)= (/ x0,l /)
!$OMP END CRITICAL
            endif
        endif
      end do
    end do
  end do
!$OMP end do
!$OMP end parallel

if (punti .eq. gp) print*, 'Good points treshold exceeded. There are ', effp - punti,' missing points.'
if (punti2 .eq. rp) print*, 'Rough points treshold exceeded. There are ',effp2 - punti2,' missing points.'

  open(unit=3,file='./tmp/disegno.dat',form='formatted',status='new',asynchronous='yes')
  open(unit=4,file='./tmp/disegno2.dat',form='formatted',status='new',asynchronous='yes')
!$OMP parallel sections private(i) num_threads(2)
!$OMP section
  do i=1,punti
    write(3,*) real(punti_buoni(:,i)), real(sqrt(punti_buoni(1,i)**2+punti_buoni(2,i)**2))
  end do 
!$OMP section
  do i=1,punti2
    write(4,*) real(punti_grezzi(:,i)), real(sqrt(punti_grezzi(1,i)**2+punti_grezzi(2,i)**2))
  end do 
!$OMP end parallel sections
  close(3)
  close(4)
  print*, punti,'/',punti2,' good/rough points recorded.'
  
  deallocate(punti_grezzi)
  deallocate(punti_buoni)
  

  end program disegno

