module service  !! contains all procedures and common data
  use rungekutta
  implicit none
  !integer,parameter :: db = kind(1.d0)
  integer,parameter :: dbl = selected_real_kind(p=33,r=200)
  real(kind=db),parameter :: overlap = 0.65   !! 30% overlap of nearby cells: 0.5= no overlap
  real(kind=db),parameter :: damping = 0.98   !! damping for the Newton step
  real(kind=db),parameter :: wsearch = 2.5    !! widen search to neighboring cells 
  real(kind=db),parameter :: lsrcfact = .5    !! linear search factor 
  integer,parameter :: maxcnt = 14   !! maximum number of Newton iterations
  integer,parameter :: maxcnint = 3  !! maximum number of linear searches
  
  real(kind=db) :: tempo,tol,atoll,rtoll !!
  real(kind=db),dimension(2) :: maxst    !! common to all procedures
  
  
  contains 
  
subroutine newtit(f,l,x0,y0,success)
  implicit none
  interface
    pure function f(x,l)
      integer,parameter :: db = kind(1.d0)
      real(kind=db), dimension(2) :: f
      real(kind=db), intent(IN) :: l
      real(kind=db), dimension(2), intent(IN) :: x
      end function f
  end interface  
  real(kind=db), intent (IN) :: l
  real(kind=db),dimension(2), intent (INOUT) :: x0
  real(kind=db),dimension(2), intent (IN) :: y0
  logical, intent (OUT) :: success
  
  real(kind=db),dimension(2,2) :: inv 
  real(kind=db),dimension(2):: x, y
  real(kind=dbl),dimension(2):: xt, yt, dNewt
  real(kind=db) :: steplen
  integer :: cnint,cnt
  logical :: flag0, flag1
  
  x=x0
  y=y0
  newtonit: do cnt=1,maxcnt ! Newton iteration: search for a zero in each cell
    if (all( abs(y) .lt. tol ) ) then
      success = .true.
      x0 = real(x,kind=db)
      return !! zero found, cycle aborted
    else
      call invJac(f,x,l,inv,flag0,flag1)
      if (flag0 .eqv. .true.) exit newtonit
        if (flag1 .eqv. .true.) then
          x = real(real(x,kind=dbl) - real(damping*matmul(inv, y),kind=dbl))
          y = f(x,l)
        else    
          cnint = 0
          steplen = damping
          dNewt = real(matmul(inv, y),kind=dbl)
          linsrc: do
            xt = real(x,kind=db) - steplen*dNewt
            yt = f(real(xt,kind=db),l)
            if (all(abs(yt) .lt. (1._db-1.e-4_db*steplen)*abs(y))) then
              x = xt
              y = yt
              exit linsrc
            else
              if (cnint .eq. maxcnint) then
                x = xt
                y = yt
                exit linsrc
                endif
              steplen = lsrcfact*steplen
              cnint = cnint + 1
              endif 
            end do linsrc  
          endif
      if ( any(abs(x-x0) .gt. maxst ) )  exit newtonit
      end if 
    end do newtonit
  success = .false.  
  end subroutine newtit        
          
          
  
pure function ff(xi,lambda) !! Translation operator at time=tt minus identity
   implicit none
   interface
     pure function deriv(t,lambda,xi)
       integer,parameter :: db = kind(1.d0)
       real(kind=db),intent(in) :: t,lambda
       real(kind=db),intent(in),dimension(2) :: xi
       real(kind=db),dimension(2) :: deriv
     end function deriv
   end interface
   real(kind=db), dimension(2) :: ff
   real(kind=db), intent(IN) :: lambda
   real(kind=db), dimension(2),intent(IN) :: xi
   
   ff= real(RKDP_adaptive(deriv,xi,tempo,lambda,atoll,rtoll),kind=dbl) - real(xi,kind=dbl)
end function ff


pure function fb(xi,lambda) !! Translation operator at time=tt minus identity
   implicit none
   interface
     pure function deriv(t,lambda,xi)
       integer,parameter :: db = kind(1.d0)
       real(kind=db),intent(in) :: t,lambda
       real(kind=db),intent(in),dimension(2) :: xi
       real(kind=db),dimension(2) :: deriv
     end function deriv
   end interface
   real(kind=db), dimension(2) :: fb
   real(kind=db), intent(IN) :: lambda
   real(kind=db), dimension(2),intent(IN) :: xi
   
   fb= real(RKDP_adaptive(backward,xi,tempo,lambda,atoll,rtoll),kind=dbl) - real(xi,kind=dbl)
   
   end function fb

   
pure function backward(t,lambda,xi)
  implicit none
  interface
    pure function deriv(t,lambda,xi)
      integer,parameter :: db = kind(1.d0)
      real(kind=db),intent(in) :: t,lambda
      real(kind=db),intent(in),dimension(2) :: xi
      real(kind=db),dimension(2) :: deriv
    end function deriv
  end interface
  real(kind=db),intent(in) :: t, lambda
  real(kind=db),dimension(2),intent(in) :: xi
  real(kind=db),dimension(2) :: backward
      
  backward = -deriv(-t,lambda,xi)
  end function backward
  

  subroutine invJac(funz,x,l,J,rankZero,rankOne) !!Computes the inverse of the Jacobian
  implicit none
  interface
    pure function funz(x,l)
      integer,parameter :: db  = kind(1.d0)
      real(kind=db),intent(IN),dimension(2) :: x 
      real(kind=db),intent(IN) :: l
      real(kind=db),dimension(2) :: funz
    end function funz
  end interface
  
  real(kind=db),parameter :: es = epsilon(1._db)**(1._db/3._db),&
                             nzero = 0.01_db*epsilon(1._db)
  real(kind=db),parameter,dimension(2,2) :: ep = reshape((/ es, 0._db, &
                                                            0._db, es/), (/2,2/))

  real(kind=db),dimension(2),intent(IN) :: x
  real(kind=db),intent(IN) :: l
  real(kind=db),dimension(2,2),intent(OUT) :: J
  logical,intent(OUT) :: rankZero,rankOne
  real(kind=dbl),dimension(2,2) :: JJ
  real(kind=dbl) :: det, deps 
  real(kind=db),dimension(2,2) :: epa
  real(kind=db),dimension(2,2),volatile :: temp
  
  temp = ep
  epa = temp * max(1._db,maxval(abs(x)))
  deps = 2._db*epa(1,1)
  
  rankOne = .false.
  JJ(:,1)= real(funz(x+epa(:,1),l),kind=dbl)-real(funz(x-epa(:,1),l),kind=dbl)
  JJ(:,2)= real(funz(x+epa(:,2),l),kind=dbl)-real(funz(x-epa(:,2),l),kind=dbl)
  if (all(abs(JJ) .lt. nzero)) then
    rankZero = .true.   !rank = 0
  else 
    rankZero = .false.
    det = (JJ(1,1)*JJ(2,2)-JJ(2,1)*JJ(1,2))/deps
    if (abs(det) .lt. nzero) then
      rankOne = .true.  !rank = 1
      JJ = (transpose(JJ)/sum(JJ**2))*deps
    else
      JJ = reshape( (/ JJ(2,2) , -JJ(2,1), -JJ(1,2), JJ(1,1) /), shape(JJ))/det
    endif
  endif
  J = real(JJ)
end subroutine invJac

end module service
    